=== modified file 'NEWS'
--- NEWS	2006-10-13 10:50:33 +0000
+++ NEWS	2006-10-17 13:58:01 +0000
@@ -40,7 +40,12 @@
   BUG FIXES:
 
     * Give nicer error message when an http server returns a 403
-      error code. (Vincent Ladeuil, # 57644).
+      error code. (Vincent Ladeuil, #57644).
+
+    * When a multi-range http GET request fails, try a single
+      range one. If it fails too, forget about ranges. Remember that until 
+      the death of the transport and propagates that to the clones.
+      (Vincent Ladeuil, #62276, #62029).
 
     * Don't require ``Content-Type`` in range responses. Assume they are a
       single range if ``Content-Type`` does not exist.

=== modified file 'bzrlib/tests/test_http.py'
--- bzrlib/tests/test_http.py	2006-10-13 15:53:16 +0000
+++ bzrlib/tests/test_http.py	2006-10-17 13:58:01 +0000
@@ -302,7 +302,7 @@
         url = '%s://%s:%s/' % (scheme, server.host, server.port)
         try:
             http_transport = get_transport(url)
-        except UnsupportedProtocol:
+        except errors.UnsupportedProtocol:
             raise TestSkipped('%s not available' % scheme)
         code, response = http_transport._post('abc def end-of-body')
         self.assertTrue(
@@ -530,16 +530,14 @@
 
     def setUp(self):
         TestCaseWithWebserver.setUp(self)
-        self.build_tree(['0123456789',],
-                        line_endings='binary',
-                        transport=self.get_transport())
+        self.build_tree_contents([('a', '0123456789')],)
 
     """Tests readv requests against server"""
 
     def test_readv(self):
         server = self.get_readonly_server()
         t = self._transport(server.get_url())
-        l = list(t.readv('0123456789', ((0, 1), (1, 1), (3, 2), (9, 1))))
+        l = list(t.readv('a', ((0, 1), (1, 1), (3, 2), (9, 1))))
         self.assertEqual(l[0], (0, '0'))
         self.assertEqual(l[1], (1, '1'))
         self.assertEqual(l[2], (3, '34'))
@@ -548,7 +546,7 @@
     def test_readv_out_of_order(self):
         server = self.get_readonly_server()
         t = self._transport(server.get_url())
-        l = list(t.readv('0123456789', ((1, 1), (9, 1), (0, 1), (3, 2))))
+        l = list(t.readv('a', ((1, 1), (9, 1), (0, 1), (3, 2))))
         self.assertEqual(l[0], (1, '1'))
         self.assertEqual(l[1], (9, '9'))
         self.assertEqual(l[2], (0, '0'))
@@ -561,12 +559,12 @@
         # This is intentionally reading off the end of the file
         # since we are sure that it cannot get there
         self.assertListRaises((errors.ShortReadvError, AssertionError),
-                              t.readv, '0123456789', [(1,1), (8,10)])
+                              t.readv, 'a', [(1,1), (8,10)])
 
         # This is trying to seek past the end of the file, it should
         # also raise a special error
         self.assertListRaises(errors.ShortReadvError,
-                              t.readv, '0123456789', [(12,2)])
+                              t.readv, 'a', [(12,2)])
 
 
 class TestSingleRangeRequestServer(TestRangeRequestServer):
@@ -609,4 +607,3 @@
     """Tests range requests refusing server for pycurl implementation"""
 
 
-

=== modified file 'bzrlib/transport/http/__init__.py'
--- bzrlib/transport/http/__init__.py	2006-10-13 10:50:33 +0000
+++ bzrlib/transport/http/__init__.py	2006-10-17 13:58:01 +0000
@@ -125,7 +125,7 @@
     # _proto: "http" or "https"
     # _qualified_proto: may have "+pycurl", etc
 
-    def __init__(self, base):
+    def __init__(self, base, from_transport=None):
         """Set the base path where files will be stored."""
         proto_match = re.match(r'^(https?)(\+\w+)?://', base)
         if not proto_match:
@@ -142,6 +142,16 @@
             self._path, self._parameters,
             self._query, self._fragment) = urlparse.urlparse(self.base)
         self._qualified_proto = apparent_proto
+        # range hint is handled dynamically throughout the life
+        # of the object. We start by trying mulri-range requests
+        # and if the server returns bougs results, we retry with
+        # single range requests and, finally, we forget about
+        # range if the server really can't understand. Once
+        # aquired, this piece of info is propogated to clones.
+        if from_transport is not None:
+            self._range_hint = from_transport._range_hint
+        else:
+            self._range_hint = 'multi'
 
     def abspath(self, relpath):
         """Return the full url to the given relative path.
@@ -252,12 +262,41 @@
                 relpath, len(offsets), ranges)
         code, f = self._get(relpath, ranges)
         for start, size in offsets:
-            f.seek(start, (start < 0) and 2 or 0)
-            start = f.tell()
-            data = f.read(size)
-            if len(data) != size:
-                raise errors.ShortReadvError(relpath, start, size,
-                                             actual=len(data))
+            try_again = True
+            while try_again:
+                try_again = False
+                f.seek(start, (start < 0) and 2 or 0)
+                start = f.tell()
+                try:
+                    data = f.read(size)
+                    if len(data) != size:
+                        raise errors.ShortReadvError(relpath, start, size,
+                                                     actual=len(data))
+                except (errors.InvalidRange, errors.ShortReadvError):
+                    # The server does not gives us enough data or
+                    # bogus-looking result, let's try again with
+                    # a simpler request if possible.
+                    if self._range_hint == 'multi':
+                        self._range_hint = 'single'
+                        mutter('Retry %s with single range request' % relpath)
+                        try_again = True
+                    elif self._range_hint == 'single':
+                        self._range_hint = None
+                        mutter('Retry %s without ranges' % relpath)
+                        try_again = True
+                    if try_again:
+                        # Note that since the offsets and the
+                        # ranges may not be in the same order we
+                        # dont't try to calculate a restricted
+                        # single range encompassing unprocessed
+                        # offsets. Note that we replace 'f' here
+                        # and that it may need cleaning one day
+                        # before being thrown that way.
+                        code, f = self._get(relpath, ranges)
+                    else:
+                        # We tried all the tricks, nothing worked
+                        raise
+
             yield start, data
 
     @staticmethod
@@ -395,14 +434,43 @@
         else:
             return self.__class__(self.abspath(offset), self)
 
+    def attempted_range_header(self, ranges, tail_amount):
+        """Prepare a HTTP Range header at a level the server should accept"""
+
+        if self._range_hint == 'multi':
+            # Nothing to do here
+            return self.range_header(ranges, tail_amount)
+        elif self._range_hint == 'single':
+            # Combine all the requested ranges into a single
+            # encompassing one
+            if len(ranges) > 0:
+                start, ignored = ranges[0]
+                ignored, end = ranges[-1]
+                if tail_amount not in (0, None):
+                    # Nothing we can do here to combine ranges
+                    # with tail_amount, just returns None. The
+                    # whole file should be downloaded.
+                    return None
+                else:
+                    return self.range_header([(start, end)], 0)
+            else:
+                # Only tail_amount, requested, leave range_header
+                # do its work
+                return self.range_header(ranges, tail_amount)
+        else:
+            return None
+
     @staticmethod
     def range_header(ranges, tail_amount):
         """Turn a list of bytes ranges into a HTTP Range header value.
 
-        :param offsets: A list of byte ranges, (start, end). An empty list
-        is not accepted.
+        :param ranges: A list of byte ranges, (start, end).
+        :param tail_amount: The amount to get from the end of the file.
 
         :return: HTTP range header string.
+
+        At least a non-empty ranges *or* a tail_amount must be
+        provided.
         """
         strings = []
         for start, end in ranges:

=== modified file 'bzrlib/transport/http/_pycurl.py'
--- bzrlib/transport/http/_pycurl.py	2006-10-13 10:50:33 +0000
+++ bzrlib/transport/http/_pycurl.py	2006-10-17 13:58:01 +0000
@@ -73,9 +73,8 @@
 
     PyCurl is a Python binding to the C "curl" multiprotocol client.
 
-    This transport can be significantly faster than the builtin Python client. 
-    Advantages include: DNS caching, connection keepalive, and ability to 
-    set headers to allow caching.
+    This transport can be significantly faster than the builtin
+    Python client.  Advantages include: DNS caching.
     """
 
     def __init__(self, base, from_transport=None):
@@ -185,7 +184,12 @@
         curl = self._range_curl
         abspath, data, header = self._setup_get_request(curl, relpath)
 
-        curl.setopt(pycurl.RANGE, self.range_header(ranges, tail_amount))
+        range_header = self.attempted_range_header(ranges, tail_amount)
+        if range_header is None:
+            # Forget ranges, the server can't handle them
+            return self._get_full(relpath)
+
+        curl.setopt(pycurl.RANGE, range_header)
         self._curl_perform(curl)
         data.seek(0)
 

=== modified file 'bzrlib/transport/http/_urllib.py'
--- bzrlib/transport/http/_urllib.py	2006-10-13 10:50:33 +0000
+++ bzrlib/transport/http/_urllib.py	2006-10-17 13:58:01 +0000
@@ -45,13 +45,11 @@
         """Set the base path where files will be stored."""
         super(HttpTransport_urllib, self).__init__(base)
         if from_transport is not None:
-            self._accept_ranges = from_transport._accept_ranges
             self._connection = from_transport._connection
             self._user = from_transport._user
             self._password = from_transport._password
             self._opener = from_transport._opener
         else:
-            self._accept_ranges = True
             self._connection = None
             self._user = None
             self._password = None
@@ -123,8 +121,10 @@
         abspath = self._real_abspath(relpath)
         headers = {}
         if ranges or tail_amount:
-            bytes = 'bytes=' + self.range_header(ranges, tail_amount)
-            headers = {'Range': bytes}
+            range_header = self.attempted_range_header(ranges, tail_amount)
+            if range_header is not None:
+                bytes = 'bytes=' + range_header
+                headers = {'Range': bytes}
 
         request = Request('GET', abspath, None, headers)
         response = self._perform(request)

